package main

import (
	"fmt"
	"os"
	"path/filepath"
	"strings"

	"go.wit.com/dev/alexflint/arg"
	"go.wit.com/gui"
	"go.wit.com/lib/gui/repolist"
	"go.wit.com/lib/gui/repostatus"
	"go.wit.com/lib/gui/shell"
	"go.wit.com/log"
)

// sent via -ldflags
var VERSION string
var BUILDTIME string

var rv *repolist.RepoList

func main() {
	pp := arg.MustParse(&argv)

	// for very new users or users unfamilar with the command line, this may help them
	if argv.Repo == "version" || argv.Repo == "help" || argv.Repo == "?" {
		pp.WriteHelp(os.Stdout)
		os.Exit(0)
	}

	// use ~/go/src unless we find a go.work file in a parent directory
	wdir, err := findWorkFile()
	if err != nil {
		log.Info(err)
		os.Exit(-1)
	}
	if wdir != "" {
		os.Setenv("REPO_WORK_PATH", wdir)

	}

	if argv.Repo == "" {
		// if there isn't anything else, just exit here
		// if --git-pull, continue
		if argv.Pull || argv.RedoGoMod {
			// there is more to do
			log.Info("onward and upward")
		} else {
			// user needs to pick something to do
			pp.WriteHelp(os.Stdout)
			log.Info("give me something to do!")
			os.Exit(0)
		}
	} else {
		// the user specified a repo, check if it's already cloned
		fullgitdir := filepath.Join(wdir, argv.Repo, ".git")
		if shell.IsDir(fullgitdir) {
			// if --recursive, continue
			// if --git-pull, continue
			if argv.Recursive || argv.Pull || argv.RedoGoMod {
				log.Info("repo already cloned", filepath.Join(wdir, argv.Repo))
				// there is more to do
				log.Info("argv.Recursive is", argv.Recursive)
				log.Info("argv.Pull is", argv.Pull)
				log.Info("argv.RedoGoMod is", argv.RedoGoMod)
				log.Info("onward and upward")
			} else {
				log.Info("repo already cloned", filepath.Join(wdir, argv.Repo))
				os.Exit(0)
			}
		} else {
			// need to download this new repo!
			log.Info("repo is new. going to clone it to:", filepath.Join(wdir, argv.Repo))
		}
	}

	// sets up gui stuff. not working yet
	b := gui.RawBox()
	rv = repolist.AutotypistView(b)

	// if the user defined a repo, attempt to download it now
	if argv.Repo != "" {
		os.Setenv("REPO_AUTO_CLONE", "true")
		newr, err := rv.NewRepo(argv.Repo)
		if err != nil {
			log.Info("could not download:", err)
			os.Exit(-1)
		}
		if argv.Recursive || argv.Pull || argv.RedoGoMod {
			log.Info("repo already cloned", filepath.Join(wdir, argv.Repo))
			// there is more to do
			log.Info("argv.Recursive is", argv.Recursive)
			log.Info("argv.Pull is", argv.Pull)
			log.Info("argv.RedoGoMod is", argv.RedoGoMod)
			log.Info("onward and upward")
		} else {
			log.Info("repo cloned worked to", filepath.Join(wdir, argv.Repo))
			os.Exit(0)
		}

		// update go.sum and go.mod
		// todo: only do this if they don't exist?
		// todo: make these git commit metadata
		newr.Status.MakeRedomod()

		// double check it actually downloaded
		fullgitdir := filepath.Join(wdir, argv.Repo, ".git")
		if !shell.IsDir(fullgitdir) {
			log.Info("repo cloned failed", filepath.Join(wdir, argv.Repo))
			os.Exit(-1)
		}
	}

	os.Setenv("REPO_AUTO_CLONE", "false")
	// look recursively in your working directory for git repos
	totalcount := scanForRepos(wdir)

	// if --git-pull, run git pull on everything here
	if argv.Pull {
		log.Info("Total repositories:", totalcount)
		log.Info("Going to run git pull in each one")
		log.Sleep(1)
		pull := []string{"git", "pull"}

		var trycount, errcount int
		loop := rv.ReposSortByName()
		for loop.Scan() {
			repo := loop.Repo()
			if argv.DryRun {
				log.Info("git pull --dry-run", repo.Status.Path())
			} else {
				trycount += 1
				log.Info("actually run: git pull:", repo.Status.Path())
				if result := shell.PathRunRealtime(repo.Status.Path(), pull); result.Error != nil {
					log.Info("git pull error:", result.Error)
					errcount += 1
				}
			}
		}
		log.Info("Total repositories:", totalcount, "Total attempted:", trycount, "Errors:", errcount)
		os.Exit(0)
	}

	// this is experiemental but works for me
	if argv.Recursive {
		newr := rv.FindRepoByName(argv.Repo)
		if newr == nil {
			log.Info("how did this repo still not exist?", argv.Repo)
			os.Exit(-1)
		}
		os.Setenv("REPO_AUTO_CLONE", "true")
		godep := newr.Status.GetGoDeps()
		for gopath, version := range godep {
			repo, err := rv.NewRepo(gopath)
			if err != nil {
				log.Info("git clone failed for", gopath, version)
				continue
			}
			// always do this for now. probably always forever
			repo.Status.MakeRedomod()
		}
	}

	// remake all the go.sum & go.mod in every repo
	// todo: make go.sum and go.mod git commit metadata
	if argv.RedoGoMod {
		loop := rv.ReposSortByName()
		for loop.Scan() {
			repo := loop.Repo()
			repo.Status.MakeRedomod()
		}
	}

	// remake the go.work file
	if argv.AutoWork {
		log.Info("About to re-create", wdir+"/go.work")
		log.Info("Sleep 3. original go.work saved as go.work.last (hit ctrl-c to cancel)")
		log.Sleep(3)
		shell.PathRun(wdir, []string{"mv", "go.work", "go.work.last"})
		rv.MakeGoWork()
		shell.PathRun(wdir, []string{"go", "work", "use"})
		log.Info("")
		log.Info("original go.work file saved as go.work.last")
		log.Info("")
	}

	log.Info("Total repositories:", totalcount)
	log.Info("Finished go-clone for", argv.Repo)
}

// look for a go.work file
// otherwise use ~/go/src
func findWorkFile() (string, error) {
	pwd, err := os.Getwd()
	if err == nil {
		// Check for go.work in the current directory and then move up until root
		if pwd, err := digup(pwd); err == nil {
			log.Info("using go.work file in directory", pwd)
			// found an existing go.work file
			os.Chdir(pwd)
			return pwd, nil
		}
	}

	// there are no go.work files, resume the ~/go/src behavior from prior to golang 1.22
	pwd, err = useGoSrc()
	log.Info("using ~/go/src directory", pwd)
	return pwd, err
}

// this is the 'old way" and works fine for me. I use it because I like the ~/go/src directory
// because I know exactly what is in it: GO stuff & nothing else
func useGoSrc() (string, error) {
	homeDir, err := os.UserHomeDir()
	if err != nil {
		return "", err
	}
	pwd := filepath.Join(homeDir, "go/src")
	shell.Mkdir(pwd)
	os.Chdir(pwd)
	return pwd, nil
}

func digup(path string) (string, error) {
	for {
		workFilePath := filepath.Join(path, "go.work")
		if _, err := os.Stat(workFilePath); err == nil {
			return path, nil // Found the go.work file
		} else if !os.IsNotExist(err) {
			return "", err // An error other than not existing
		}

		parentPath := filepath.Dir(path)
		if parentPath == path {
			break // Reached the filesystem root
		}
		path = parentPath
	}

	return "", fmt.Errorf("no go.work file found")
}

func scanForRepos(wdir string) int {
	var count int
	log.Info("scanning for repo in:", filepath.Join(wdir, argv.Repo))

	// rv.NewRepo("go.wit.com/apps/helloworld")
	for _, path := range repostatus.ScanGitDirectories(wdir) {
		count += 1
		gopath := strings.TrimPrefix(path, wdir)
		gopath = strings.Trim(gopath, "/")
		// log.Info("Also should add:", gopath)
		rv.NewRepo(gopath)
	}
	return count
}
